"use strict";
exports.__esModule = true;
var tslib_1 = require("tslib");
var constants_1 = require("./constants");
/**
 * Scan parser context and look for container
 * matches based on expression path.
 *
 * -- Brief description --
 * An expression path is an array of strings determining a container reference,
 * either it points directly to a container type (matching various containers) or
 * it points directly to a single container.
 *
 * Context scanner will look into AST trying to find namespaces that matches the
 * given path.
 *
 * TODO: Improve searching algorithm and break it in many parts
 */
var ContextScanner = /** @class */ (function () {
    function ContextScanner(parser) {
        this.parser = parser;
    }
    ContextScanner.prototype.match = function (expressionPath) {
        var result;
        var path = this.parser.path;
        if (path.length > 0) {
            var ii = path.length;
            for (var i = ii; i >= 0; i--) {
                result = this._match(path.slice(0, i).concat(expressionPath));
                if (result.length)
                    return result;
            }
        }
        return this._match(expressionPath);
    };
    ContextScanner.prototype._match = function (expressionPath) {
        var matches = [];
        var body = this.parser.ast;
        var expressionIndex = 0;
        var path = [];
        bodyLoop: while (true) {
            for (var i = 0; i < body.length; i++) {
                var item = body[i];
                if (expressionIndex == (expressionPath.length - 1)) {
                    if (item.type != constants_1.Syntax.ContainerGroup)
                        continue;
                    if (item.name == expressionPath[expressionIndex]) {
                        // collect container declaration names
                        for (var j = 0; j < item.body.length; j++) {
                            var item2 = item.body[j];
                            if (item2.type == constants_1.Syntax.ContainerDeclaration) {
                                matches.push(path.concat([item2.name]));
                            }
                        }
                        if (matches.length)
                            return matches;
                    }
                    else {
                        for (var j = 0; j < item.body.length; j++) {
                            var item2 = item.body[j];
                            if (item2.type != constants_1.Syntax.ContainerDeclaration)
                                continue;
                            if (item2.name == expressionPath[expressionIndex])
                                return [path.concat(item2.name)];
                        }
                    }
                    continue;
                }
                if (item.type == constants_1.Syntax.Namespace && item.name == expressionPath[expressionIndex]) {
                    path.push(item.name);
                    body = item.body;
                    expressionIndex++;
                    continue bodyLoop;
                }
            }
            break;
        }
        return matches;
    };
    /**
     * Resolve a type inside a context
     * @param name Container name/container type to look for
     * @param body Body to look for
     * @param namespace Namespace to start from
     */
    ContextScanner.prototype.resolveTypeInBody = function (name, body, namespace) {
        var e_1, _a, e_2, _b;
        var results = new Array();
        try {
            for (var body_1 = (0, tslib_1.__values)(body), body_1_1 = body_1.next(); !body_1_1.done; body_1_1 = body_1.next()) {
                var node = body_1_1.value;
                /**
                 * Since the `name` can't point to a namespace
                 * we have to ignore if this is not a container name or container group
                 */
                if (namespace.length === 0) {
                    if (node.type !== constants_1.Syntax.ContainerGroup && node.type !== constants_1.Syntax.ContainerDeclaration) {
                        continue;
                    }
                    if (node.type === constants_1.Syntax.ContainerGroup) {
                        if (node.name === name) {
                            results.push(node);
                            continue;
                        }
                        try {
                            for (var _c = (e_2 = void 0, (0, tslib_1.__values)(node.body)), _d = _c.next(); !_d.done; _d = _c.next()) {
                                var item = _d.value;
                                if (item.type === constants_1.Syntax.TemplateDeclaration) {
                                    item = item.body;
                                }
                                if (item.type === constants_1.Syntax.ContainerDeclaration && item.name === name) {
                                    results.push(item);
                                    break;
                                }
                            }
                        }
                        catch (e_2_1) { e_2 = { error: e_2_1 }; }
                        finally {
                            try {
                                if (_d && !_d.done && (_b = _c["return"])) _b.call(_c);
                            }
                            finally { if (e_2) throw e_2.error; }
                        }
                    }
                    else if (name === node.name) {
                        results.push(node);
                    }
                    continue;
                }
                else if (node.type === constants_1.Syntax.Namespace) {
                    if (node.name === namespace[0]) {
                        results.push.apply(results, (0, tslib_1.__spreadArray)([], (0, tslib_1.__read)(this.resolveTypeInBody(name, node.body, namespace.slice(1))), false));
                        continue;
                    }
                }
            }
        }
        catch (e_1_1) { e_1 = { error: e_1_1 }; }
        finally {
            try {
                if (body_1_1 && !body_1_1.done && (_a = body_1["return"])) _a.call(body_1);
            }
            finally { if (e_1) throw e_1.error; }
        }
        return results;
    };
    return ContextScanner;
}());
exports["default"] = ContextScanner;
