"use strict";
exports.__esModule = true;
var tslib_1 = require("tslib");
var context_scanner_1 = (0, tslib_1.__importDefault)(require("./context-scanner"));
var cyclic_rc_1 = (0, tslib_1.__importDefault)(require("cyclic-rc"));
// Default template processors
var vector_processor_1 = (0, tslib_1.__importDefault)(require("./vector-processor"));
var optional_processor_1 = (0, tslib_1.__importDefault)(require("./optional-processor"));
var typed_array_processor_1 = (0, tslib_1.__importDefault)(require("./typed-array-processor"));
var constants_1 = require("./constants");
var param_1 = require("./param");
var ast_preprocessor_1 = (0, tslib_1.__importDefault)(require("./ast-preprocessor"));
var strict_size_processor_1 = (0, tslib_1.__importDefault)(require("./strict-size-processor"));
var utils_1 = require("./utils");
var map_processor_1 = (0, tslib_1.__importDefault)(require("./map-processor"));
var ASTParser = /** @class */ (function () {
    function ASTParser(ast, options) {
        ASTParser.defineDefaultOptions(options);
        this.ast = new ast_preprocessor_1["default"](ast, {
            directory: options.directory
        }).getResult();
        this.path = options.path || [];
        this.parent = options.parent;
        this.scanner = new context_scanner_1["default"](this);
        this.options = options;
        this.containers = options.containers;
        this.namespaceSeparator = options.namespaceSeparator || '::';
        this.templateProcessors = new Map();
        this.templateProcessors.set('Vector', new vector_processor_1["default"](this));
        this.templateProcessors.set('Optional', new optional_processor_1["default"](this));
        this.templateProcessors.set('Map', new map_processor_1["default"](this));
        this.templateProcessors.set('TypedArray', new typed_array_processor_1["default"](this));
        this.templateProcessors.set('StrictSize', new strict_size_processor_1["default"](this));
    }
    ASTParser.defineDefaultOptions = function (options) {
        var defaults = {
            namespaceSeparator: '::',
            path: []
        };
        Object.keys(defaults).forEach(function (prop) {
            if (!options.hasOwnProperty(prop))
                options[prop] = defaults[prop];
        });
    };
    ASTParser.prototype.parse = function () {
        var e_1, _a;
        try {
            for (var _b = (0, tslib_1.__values)(this.ast), _c = _b.next(); !_c.done; _c = _b.next()) {
                var item = _c.value;
                this.parseAstItem(item);
            }
        }
        catch (e_1_1) { e_1 = { error: e_1_1 }; }
        finally {
            try {
                if (_c && !_c.done && (_a = _b["return"])) _a.call(_b);
            }
            finally { if (e_1) throw e_1.error; }
        }
    };
    ASTParser.prototype.parseParamType = function (paramType) {
        var _this = this;
        switch (paramType.type) {
            case constants_1.Syntax.Identifier: {
                var type = paramType.value;
                if ((0, utils_1.isGeneric)(type)) {
                    return {
                        type: param_1.Params.Generic,
                        name: type
                    };
                }
                var matches = this.scanner.match([type]);
                if (matches.length == 0)
                    throw new Error("No match found for type \"".concat(type, "\""));
                return {
                    type: param_1.Params.Reference,
                    containers: matches.map(function (m) { return m.join(_this.namespaceSeparator); })
                };
            }
            case constants_1.Syntax.Template: {
                var templateProcessor = this.templateProcessors.get(paramType.name);
                var outputArguments = void 0;
                if (!templateProcessor) {
                    var results = this.scanner.resolveTypeInBody(paramType.name, this.ast, this.path);
                    if (results.length === 0) {
                        throw new Error("Unexpected template variable \"".concat(paramType.name, "\""));
                    }
                    outputArguments = paramType.arguments;
                }
                else {
                    outputArguments = templateProcessor.process(paramType.arguments);
                }
                return {
                    type: param_1.Params.Template,
                    name: paramType.name,
                    arguments: outputArguments
                };
            }
            case constants_1.Syntax.MemberExpression: {
                var expressionPath = this.processMemberExpression(paramType);
                var matches = this.scanner.match(expressionPath);
                if (matches.length == 0)
                    throw new Error("No match found for member expression ".concat(expressionPath.join(' -> ')));
                return {
                    type: param_1.Params.Reference,
                    containers: matches.map(function (m) { return m.join(_this.namespaceSeparator); })
                };
            }
        }
        throw new Error("Unexpected param type \"".concat(paramType.type, "\""));
    };
    ASTParser.prototype.processMemberExpression = function (paramType) {
        if (paramType.type == constants_1.Syntax.MemberExpression) {
            var left = this.processMemberExpression(paramType.left);
            return (0, tslib_1.__spreadArray)((0, tslib_1.__spreadArray)([], (0, tslib_1.__read)(left), false), [paramType.right], false);
        }
        else if (paramType.type == constants_1.Syntax.Identifier) {
            return [paramType.value];
        }
        else {
            throw new Error('Invalid member expression');
        }
    };
    ASTParser.prototype.parseContainerParam = function (param) {
        return {
            name: param.name,
            type: this.parseParamType(param.paramType)
        };
    };
    ASTParser.prototype.createCrcString = function (ast) {
        var _this = this;
        switch (ast.type) {
            case constants_1.Syntax.ContainerDeclaration:
                return "".concat(this.path.concat([ast.name]).join('.'), " -> ") +
                    "".concat(ast.body.map(function (node) { return _this.createCrcString(node); }).join(','));
            case constants_1.Syntax.ContainerParam:
                return "".concat(ast.name, ":").concat(this.createCrcString(ast.paramType));
            case constants_1.Syntax.Identifier:
                return ast.value;
            case constants_1.Syntax.MemberExpression:
                return "".concat(this.createCrcString(ast.left), ".").concat(ast.right);
            case constants_1.Syntax.Template: {
                var processor = this.templateProcessors.get(ast.name);
                var args = void 0;
                if (!processor) {
                    var results = this.scanner.resolveTypeInBody(ast.name, this.ast, this.path);
                    if (results.length === 0) {
                        throw new Error("No template processor found to process ".concat(ast.name, " arguments"));
                    }
                    args = ast.arguments.map(function (arg) { return _this.createCrcString(arg); });
                }
                else {
                    args = processor.createCrcString(ast.arguments);
                }
                return "".concat(ast.name, "<").concat(args.join(','), ">");
            }
        }
        throw new Error("Unexpected node type for crc string generation: ".concat(ast.type));
    };
    ASTParser.prototype.processContainerGroupNode = function (ast, parent) {
        var e_2, _a;
        switch (ast.type) {
            case constants_1.Syntax.ContainerDeclaration: {
                var params = [];
                try {
                    for (var _b = (0, tslib_1.__values)(ast.body), _c = _b.next(); !_c.done; _c = _b.next()) {
                        var item = _c.value;
                        if (item.type == constants_1.Syntax.ContainerParam) {
                            params.push(this.parseContainerParam(item));
                        }
                    }
                }
                catch (e_2_1) { e_2 = { error: e_2_1 }; }
                finally {
                    try {
                        if (_c && !_c.done && (_a = _b["return"])) _a.call(_b);
                    }
                    finally { if (e_2) throw e_2.error; }
                }
                var crcValue = "".concat(parent.name, " ").concat(this.createCrcString(ast));
                this.containers.push({
                    id: cyclic_rc_1["default"].crc_32(crcValue),
                    type: this.path.concat([parent.name]).join(this.namespaceSeparator),
                    params: params,
                    name: this.path.concat([ast.name]).join(this.namespaceSeparator)
                });
                return;
            }
            case constants_1.Syntax.TemplateDeclaration:
                return;
        }
        throw new Error("Invalid node for container group \"".concat(ast.type, "\""));
    };
    ASTParser.prototype.parseAstItem = function (ast) {
        var e_3, _a, e_4, _b;
        switch (ast.type) {
            case constants_1.Syntax.Namespace: {
                this.path.push(ast.name);
                try {
                    for (var _c = (0, tslib_1.__values)(ast.body), _d = _c.next(); !_d.done; _d = _c.next()) {
                        var item = _d.value;
                        this.parseAstItem(item);
                    }
                }
                catch (e_3_1) { e_3 = { error: e_3_1 }; }
                finally {
                    try {
                        if (_d && !_d.done && (_a = _c["return"])) _a.call(_c);
                    }
                    finally { if (e_3) throw e_3.error; }
                }
                this.path.pop();
                break;
            }
            case constants_1.Syntax.ContainerGroup: {
                var body = ast.body;
                try {
                    for (var body_1 = (0, tslib_1.__values)(body), body_1_1 = body_1.next(); !body_1_1.done; body_1_1 = body_1.next()) {
                        var item = body_1_1.value;
                        this.processContainerGroupNode(item, ast);
                    }
                }
                catch (e_4_1) { e_4 = { error: e_4_1 }; }
                finally {
                    try {
                        if (body_1_1 && !body_1_1.done && (_b = body_1["return"])) _b.call(body_1);
                    }
                    finally { if (e_4) throw e_4.error; }
                }
                break;
            }
        }
    };
    return ASTParser;
}());
exports["default"] = ASTParser;
