import { Node, NodeContainerDeclaration, NodeContainerGroup, NodeMemberExpression, NodeIdentifier } from '../ast-parser/node';
import { Container } from '../ast-parser/constants';
import { ASTParserOptions } from '../ast-parser';
import ModuleResolver from './module-resolver';
export interface IFileResult {
    path: string;
    contents: string;
    namespace: string[];
}
export interface CodeGeneratorOptions extends Partial<ASTParserOptions> {
    prefixPackage: string[];
    indentationSize: number;
    moduleAliases: Map<string, string>;
}
export interface ICodeGenerator<T = string> {
    getFiles(): IFileResult[];
    setParentNode(node?: Node): void;
    processNode(node: Node): string;
    getVariableName(name: string, variant?: string, includeComment?: boolean): string;
    indentCode(...args: string[]): string;
    getPackageName(): string;
    increaseDepth(): void;
    decreaseDepth(): void;
    generate(node: Node): T;
    getContainers(): Container[];
    getModuleResolver(): ModuleResolver;
    getContainer(name: string): Container;
    getNodeListFromParamExpression(node: Node): Node[];
    getInterfaceName(name: string): string;
    getCurrentNode(): Node | undefined;
    setCurrentNode(node?: Node): void;
    getParamDefaultFromType(node: Node): string;
    /**
     * Returns the node that holds the current node
     */
    getParentNode(): Node | undefined;
    translateParamType(node: Node): string;
    getPrefixPackage(): string[];
    /**
     * Return current namespace path we're in
     */
    getCurrentPath(): string[];
    addFile(file: Pick<IFileResult, 'path' | 'contents'>): void;
    getClassNameFromList(node: NodeContainerDeclaration | NodeContainerGroup): string;
    getGenerator<GeneratorType extends ICodeGenerator>(name: string): GeneratorType;
}
export default class CodeGenerator implements ICodeGenerator {
    private nodes;
    generators: Map<string, ICodeGenerator<string>>;
    private files;
    private path;
    private depth;
    private currentNode?;
    private parentNode?;
    private prefixPackage;
    private containers;
    private indentationSize;
    private moduleResolver;
    constructor(nodes: Node[], options?: Partial<CodeGeneratorOptions>);
    getModuleResolver(): ModuleResolver;
    getGenerator<T extends ICodeGenerator = ICodeGenerator>(name: string): T;
    getFiles(): IFileResult[];
    addFile(file: Pick<IFileResult, 'path' | 'contents'>): void;
    getCurrentNode(): Node | undefined;
    getPrefixPackage(): string[];
    getParentNode(): Node | undefined;
    increaseDepth(): void;
    decreaseDepth(): void;
    indentCode(...args: string[]): string;
    getParamDefaultFromType(node: Node): string;
    getContainers(): Container[];
    setParentNode(node?: Node): void;
    processNode(item: Node): string;
    setCurrentNode(node?: Node): void;
    /**
     * Generates an unique identifier according to the current parameter, namespace and container name. It is
     * safe to use this function and it'll always return the same value as long as you keep inside container's context.
     *
     * As you already know it's safety, you should always use `variant` to specify some specific information for that
     * variable whenever you'll be dealing with a different value and don't need that specific value to be overwritten
     * or to be conflicted with other variable identifiers
     *
     * TODO: check if worth the heavy work that needs to be executed so we can maintain the uniqueness of the returned
     * value. Other options are: use an incremental id and the difference will not be much as long as we keep the
     * same order of variables
     *
     * @param paramName current param name
     * @param variant optional parameter used to not generated the same hash, it should be used whenever the variable
     * is not carrying the parameter value
     */
    getVariableName(paramName: string, variant?: string, includeComment?: boolean): string;
    getContainer(name: string): Container;
    getClassNameFromList(node: Node): string;
    getNodeListFromParamExpression(node: Node): Node[];
    getNamespaceFromMemberExpression(node: NodeIdentifier | NodeMemberExpression, namespace?: string[]): string[];
    /**
     * Translate param type to a valid TypeScript type
     * @param node Node to be translated from
     */
    translateParamType(node: Node): string;
    /**
     * Resolve a type inside a context
     * @param name Container name/container type to look for
     * @param body Body to look for
     * @param namespace Namespace to start from
     */
    resolveTypeInBody(name: string, body: Node[], namespace: string[]): Node[];
    getCurrentPath(): string[];
    /**
     * Get package name according to current path
     */
    getPackageName(): string;
    getInterfaceName(name: string): string;
    generate(): string;
    private getDefaultGenerators;
}
