"use strict";
exports.__esModule = true;
var tslib_1 = require("tslib");
var constants_1 = require("../ast-parser/constants");
var param_1 = require("../ast-parser/param");
var serializer_1 = (0, tslib_1.__importDefault)(require("./serializer"));
var deserializer_1 = (0, tslib_1.__importDefault)(require("./deserializer"));
var strict_size_processor_1 = (0, tslib_1.__importDefault)(require("./strict-size-processor"));
var vector_processor_1 = (0, tslib_1.__importDefault)(require("./vector-processor"));
var optional_processor_1 = (0, tslib_1.__importDefault)(require("./optional-processor"));
var typed_array_processor_1 = (0, tslib_1.__importDefault)(require("./typed-array-processor"));
var map_processor_1 = (0, tslib_1.__importDefault)(require("./map-processor"));
var long_1 = (0, tslib_1.__importDefault)(require("long"));
var Schema = /** @class */ (function () {
    function Schema(containers, options) {
        var e_1, _a;
        this.containers = new Map();
        this.containersCRC = new Map();
        this.templateProcessors = new Map();
        try {
            for (var containers_1 = (0, tslib_1.__values)(containers), containers_1_1 = containers_1.next(); !containers_1_1.done; containers_1_1 = containers_1.next()) {
                var container = containers_1_1.value;
                this.containers.set(container.name, container);
                this.containersCRC.set(container.id, container);
            }
        }
        catch (e_1_1) { e_1 = { error: e_1_1 }; }
        finally {
            try {
                if (containers_1_1 && !containers_1_1.done && (_a = containers_1["return"])) _a.call(containers_1);
            }
            finally { if (e_1) throw e_1.error; }
        }
        this.templateProcessors.set('StrictSize', new strict_size_processor_1["default"](this));
        this.templateProcessors.set('Vector', new vector_processor_1["default"](this));
        this.templateProcessors.set('TypedArray', new typed_array_processor_1["default"](this));
        this.templateProcessors.set('Map', new map_processor_1["default"](this));
        this.templateProcessors.set('Optional', new optional_processor_1["default"](this));
        if (options && options.templateProcessors) {
            var iterator = options.templateProcessors.keys();
            var current = iterator.next();
            while (!current.done) {
                var processor = options.templateProcessors.get(current.value);
                if (processor) {
                    this.templateProcessors.set(current.value, processor);
                }
                current = iterator.next();
            }
        }
    }
    Schema.prototype.decodeGeneric = function (des, type) {
        switch (type) {
            case constants_1.Generics.Double:
                return des.readDouble();
            case constants_1.Generics.Float:
                return des.readFloat();
            case constants_1.Generics.Int16:
                return des.readInt16();
            case constants_1.Generics.UInt16:
                return des.readUInt16();
            case constants_1.Generics.UInt8:
                return des.readUInt8();
            case constants_1.Generics.Int8:
                return des.readInt8();
            case constants_1.Generics.Int32:
                return des.readInt32();
            case constants_1.Generics.UInt32:
                return des.readUInt32();
            case constants_1.Generics.Int64:
                return des.readInt64();
            case constants_1.Generics.UInt64:
                return des.readUInt64();
            case constants_1.Generics.String:
                return des.readBuffer(des.readUInt32()).toString('utf8');
            case constants_1.Generics.Boolean:
                return des.readBoolean();
            default:
                throw new Error("Can't decode generic of \"".concat(type, "\""));
        }
    };
    Schema.prototype.encodeGeneric = function (serializer, type, value) {
        switch (type) {
            case constants_1.Generics.Double:
                serializer.writeDouble(value);
                break;
            case constants_1.Generics.Float:
                serializer.writeFloat(value);
                break;
            case constants_1.Generics.Int16:
                serializer.writeInt16(value);
                break;
            case constants_1.Generics.UInt16:
                serializer.writeUInt16(value);
                break;
            case constants_1.Generics.UInt8:
                serializer.writeUInt8(value);
                break;
            case constants_1.Generics.Int8:
                serializer.writeInt8(value);
                break;
            case constants_1.Generics.Int32:
                serializer.writeInt32(value);
                break;
            case constants_1.Generics.UInt32:
                serializer.writeUInt32(value);
                break;
            case constants_1.Generics.String:
                serializer.writeString(value);
                break;
            case constants_1.Generics.Boolean:
                serializer.writeBoolean(value);
                break;
            case constants_1.Generics.UInt64:
                serializer.writeUInt64(value);
                break;
            case constants_1.Generics.Int64:
                serializer.writeInt64(value);
                break;
            default:
                throw new Error("Can't encode generic of \"".concat(type, "\""));
        }
    };
    /**
     * Get container name from object.
     * Generally defined as [$container_name, $params]
     */
    Schema.prototype.getContainerName = function (value) {
        return value._name;
    };
    Schema.prototype.getContainerParams = function (value) {
        return (0, tslib_1.__assign)({}, value);
    };
    Schema.prototype.createObject = function (container, params) {
        return [container.name, params];
    };
    Schema.prototype.decodeTemplate = function (deserializer, template, result, prop) {
        var processor = this.templateProcessors.get(template.name);
        if (!processor)
            throw new Error("No template processor for template \"".concat(template.name, "\""));
        processor.decode(deserializer, template.arguments, result, prop);
    };
    Schema.prototype.decode = function (deserializer) {
        if (Buffer.isBuffer(deserializer)) {
            return this.decode(new deserializer_1["default"](deserializer));
        }
        var id = deserializer.readUInt32();
        var container = this.containersCRC.get(id);
        if (!container)
            throw new Error("No container found for CRC hash of ".concat(id));
        var props = {};
        var params = container.params;
        var ii = params.length;
        for (var i = 0; i < ii; i++) {
            var param = params[i];
            var property = param.name;
            this.decodeContainerParam(deserializer, param.type, props, property);
        }
        return this.createObject(container, props);
    };
    Schema.prototype.decodeContainerParam = function (deserializer, param, result, prop) {
        switch (param.type) {
            case param_1.Params.Generic: {
                var decoded = this.decodeGeneric(deserializer, param.name);
                this.validateGeneric(param.name, decoded);
                result[prop] = decoded;
                break;
            }
            case param_1.Params.Reference:
                result[prop] = this.decode(deserializer);
                break;
            case param_1.Params.Template:
                this.decodeTemplate(deserializer, param, result, prop);
        }
    };
    Schema.prototype.encodeReference = function (serializer, containers, value) {
        var name = this.getContainerName(value);
        if (containers.indexOf(name) == -1)
            throw new Error("Expected ".concat(containers.join(' or '), " but got ").concat(name, " instead"));
        this.encode(value, serializer);
    };
    Schema.prototype.encodeTemplate = function (serializer, param, value) {
        var processor = this.templateProcessors.get(param.name);
        if (!processor)
            throw new Error("No template for encoding \"".concat(param.name, "\" template"));
        processor.encode(serializer, param.arguments, value);
    };
    /**
     * Return default value for generics
     */
    Schema.prototype.getGenericDefault = function (value) {
        switch (value) {
            case constants_1.Generics.Double:
            case constants_1.Generics.Float:
            case constants_1.Generics.Int16:
            case constants_1.Generics.UInt16:
            case constants_1.Generics.UInt8:
            case constants_1.Generics.Int8:
            case constants_1.Generics.Int32:
            case constants_1.Generics.UInt32:
            case constants_1.Generics.Int64:
            case constants_1.Generics.UInt64:
                return 0;
            case constants_1.Generics.String:
                return '';
            case constants_1.Generics.Boolean:
                return false;
        }
        throw new Error("Could not find default value for generic param");
    };
    Schema.prototype.validateGeneric = function (type, value) {
        switch (type) {
            case constants_1.Generics.Int64:
            case constants_1.Generics.UInt64:
                if (!long_1["default"].isLong(value) && typeof value !== 'number') {
                    throw new Error("Expected Long instance of number but got ".concat(typeof value, " instead"));
                }
                break;
            case constants_1.Generics.Double:
            case constants_1.Generics.Float:
            case constants_1.Generics.Int16:
            case constants_1.Generics.UInt16:
            case constants_1.Generics.UInt8:
            case constants_1.Generics.Int8:
            case constants_1.Generics.Int32:
            case constants_1.Generics.UInt32:
                if (typeof value !== 'number') {
                    throw new Error("Expected a number but got ".concat(typeof value, " instead"));
                }
                if (isNaN(value)) {
                    throw new Error('Expected a number but got a NaN value instead');
                }
                break;
            case constants_1.Generics.String:
                if (typeof value !== 'string') {
                    throw new Error("Expected a string but got ".concat(typeof value));
                }
                break;
            case constants_1.Generics.Boolean:
                if (typeof value !== 'boolean') {
                    throw new Error("Expected a boolean but got ".concat(typeof value));
                }
                break;
            default:
                throw new Error("Invalid generic type of ".concat(type));
        }
    };
    Schema.prototype.encodeContainerParam = function (serializer, param, value) {
        switch (param.type) {
            case param_1.Params.Generic:
                if (typeof value === 'undefined') {
                    value = this.getGenericDefault(param.name);
                }
                this.validateGeneric(param.name, value);
                this.encodeGeneric(serializer, param.name, value);
                break;
            case param_1.Params.Reference:
                this.encodeReference(serializer, param.containers, value);
                break;
            case param_1.Params.Template:
                this.encodeTemplate(serializer, param, value);
                break;
        }
    };
    Schema.prototype.encode = function (name, props, serializer) {
        if (typeof name == 'object') {
            if (props instanceof serializer_1["default"]) {
                serializer = props;
            }
            props = this.getContainerParams(name);
            name = this.getContainerName(name);
        }
        if (!serializer)
            serializer = new serializer_1["default"];
        var container = this.containers.get(name);
        if (!container)
            throw new Error("Unexpected container name \"".concat(name, "\""));
        var params = container.params;
        var ii = params.length;
        serializer.writeUInt32(container.id);
        for (var i = 0; i < ii; i++) {
            var param = params[i];
            this.encodeContainerParam(serializer, param.type, props[param.name]);
        }
        return serializer.getBuffer();
    };
    return Schema;
}());
exports["default"] = Schema;
