"use strict";
exports.__esModule = true;
var tslib_1 = require("tslib");
var long_1 = (0, tslib_1.__importDefault)(require("long"));
var Serializer = /** @class */ (function () {
    function Serializer(length) {
        if (length === void 0) { length = 1 * 1024 * 1024; }
        this.length = length;
        this.offset = 0;
        this.memoryGrowthAmount = this.length;
        this.buffer = new ArrayBuffer(this.length);
        this.view = Buffer.from(this.buffer);
    }
    Serializer.prototype.reset = function () {
        this.offset = 0;
    };
    Serializer.prototype.getBuffer = function () {
        return this.view.slice(0, this.offset);
    };
    Serializer.prototype.allocate = function (length) {
        if ((this.offset + length) >= this.length) {
            var oldView = this.view;
            this.length += length + this.memoryGrowthAmount;
            this.buffer = new ArrayBuffer(this.length);
            this.view = Buffer.from(this.buffer);
            oldView.copy(this.view, 0);
        }
    };
    Serializer.prototype.writeBuffer = function (source) {
        if (!Buffer.isBuffer(source))
            throw new Error('Argument must be a buffer instance');
        var length = source.byteLength;
        this.allocate(length);
        source.copy(this.view, this.offset, 0, length);
        this.offset += length;
    };
    Serializer.prototype.writeBoolean = function (n) {
        if (typeof n !== 'boolean') {
            throw new Error('Unexpected type for boolean value. It must be native true or false');
        }
        this.writeUInt8(n ? 1 : 0);
    };
    Serializer.prototype.writeInt8 = function (n) {
        this.allocate(1);
        this.view.writeInt8(n, this.offset);
        this.offset++;
    };
    Serializer.prototype.writeUInt8 = function (n) {
        this.allocate(1);
        this.view.writeUInt8(n, this.offset);
        this.offset++;
    };
    Serializer.prototype.writeUInt16 = function (n) {
        this.allocate(2);
        this.view.writeUInt16LE(n, this.offset);
        this.offset += 2;
    };
    Serializer.prototype.writeInt16 = function (n) {
        this.allocate(2);
        this.view.writeInt16LE(n, this.offset);
        this.offset += 2;
    };
    Serializer.prototype.writeUInt32 = function (n) {
        this.allocate(4);
        this.view.writeUInt32LE(n, this.offset);
        this.offset += 4;
    };
    Serializer.prototype.writeInt32 = function (n) {
        this.allocate(4);
        this.view.writeInt32LE(n, this.offset);
        this.offset += 4;
    };
    Serializer.prototype.writeUInt64 = function (n) {
        var long;
        if (typeof n === 'number') {
            long = new long_1["default"](n, 0);
        }
        else {
            long = n;
        }
        this.writeLong(long, true);
    };
    Serializer.prototype.writeInt64 = function (n) {
        var long;
        if (typeof n === 'number') {
            long = new long_1["default"](n, 0);
        }
        else {
            long = n;
        }
        this.writeLong(long, false);
    };
    Serializer.prototype.writeDouble = function (n) {
        this.allocate(8);
        this.view.writeDoubleLE(n, this.offset);
        this.offset += 8;
    };
    Serializer.prototype.writeFloat = function (n) {
        this.allocate(4);
        this.view.writeFloatLE(n, this.offset);
        this.offset += 4;
    };
    Serializer.prototype.writeString = function (n) {
        var str = Buffer.from(n, 'utf8');
        this.writeUInt32(str.byteLength);
        this.writeBuffer(str);
    };
    Serializer.prototype.writeLong = function (n, unsigned) {
        if (unsigned) {
            this.writeUInt32(n.getLowBitsUnsigned());
            this.writeUInt32(n.getHighBitsUnsigned());
        }
        else {
            this.writeInt32(n.getLowBits());
            this.writeInt32(n.getHighBits());
        }
    };
    return Serializer;
}());
exports["default"] = Serializer;
